/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "yac_interface.h"
#include "utils.h"

#define NBR_CELLS 2
#define NBR_VERTICES 4
#define YAC_RAD (0.01745329251994329576923690768489) // M_PI / 180

int main () {

   int no_of_fields = 8;

   const char * fieldName[] = { "atmos_out1",
                                "atmos_out2",
                                "atmos_out3",
                                "atmos_out4",
                                "ocean_out1",
                                "ocean_out2",
                                "ocean_out3",
                                "ocean_out4"};

  char * xml_filename = "dummy_coupling.xml";
  char * xsd_filename = "coupling.xsd";

  char * comp_name = "dummy_io";

  char * grid_name[] = { "ocean_grid",
                         "atmos_grid"};

  const int nbr_cells = NBR_CELLS;
  const int nbr_vertices = NBR_VERTICES;
  int nbr_vertices_per_cell[NBR_CELLS];

  int i, ierror;
  int comp_id;

  int cell_point_ids[2];

  int grid_ids[2];

  int global_index[NBR_CELLS];
  int cell_core_mask[NBR_CELLS];

  double * buffer_lon;
  double * buffer_lat;
  int * cell_to_vertex;

  int * cell_mask;
  int * field_id;

  MPI_Comm local_comm;

  int size, rank;

  MPI_Init (0, NULL);

  // Initialise the coupler
  yac_cinit ( xml_filename, xsd_filename );

  // Inform the coupler about what we are
  yac_cdef_comp ( comp_name, &comp_id );

  printf ( "YAC Version %s\n", yac_cget_version() );

  yac_cget_localcomm ( &local_comm, comp_id );

  MPI_Comm_rank ( local_comm, &rank );
  MPI_Comm_size ( local_comm, &size );

  printf (" %s rank %d : local size %d \n", comp_name, rank, size );

  /* Here we define two grids (although numrically identical) to mimick
     two different Output server groups */

  for ( int igrid = 0; igrid < 2; igrid++ ) {

    buffer_lon = xmalloc ( nbr_vertices * sizeof(*buffer_lon) );
    buffer_lat = xmalloc ( nbr_vertices * sizeof(*buffer_lat) );
    cell_to_vertex = xmalloc ( 3 * nbr_cells * sizeof(*cell_to_vertex) );

    /* Define vertices

                0
               / \
              / o \
             /     \
            1-------2   Eq.
             \     /
              \ o /
               \ /
                3

    */

    buffer_lon[0] =  0.0 * YAC_RAD; buffer_lat[0] =  1.1 * YAC_RAD;
    buffer_lon[1] = -1.0 * YAC_RAD; buffer_lat[1] =  0.0 * YAC_RAD;
    buffer_lon[2] =  1.0 * YAC_RAD; buffer_lat[2] =  0.0 * YAC_RAD;
    buffer_lon[3] =  0.0 * YAC_RAD; buffer_lat[3] = -1.0 * YAC_RAD;

    // Connectivity

    cell_to_vertex[0] = 0; cell_to_vertex[1] = 1; cell_to_vertex[2] = 2; // cell 1
    cell_to_vertex[3] = 1; cell_to_vertex[4] = 3; cell_to_vertex[5] = 2; // cell 2

    for ( i = 0; i < nbr_cells; ++i ) nbr_vertices_per_cell[i] = 3;

    // Define unstructured grid

    yac_cdef_grid_unstruct ( grid_name[igrid],
                             nbr_vertices,
                             nbr_cells,
                             nbr_vertices_per_cell,
                             buffer_lon,
                             buffer_lat,
                             cell_to_vertex,
                             &grid_ids[igrid] );

  // Decomposition information

  for (int i = 0; i < nbr_cells; ++i ) {
    global_index[i] = i;
    cell_core_mask[i] = 1;
  }

  yac_cset_global_index(global_index, YAC_LOCATION_CELL, grid_ids[igrid]);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, grid_ids[igrid]);

    // Center points in cells (needed e.g. for nearest neighbour)

    buffer_lon[0] = 0.0 * YAC_RAD; buffer_lat[0] =  0.5 * YAC_RAD;
    buffer_lon[1] = 0.0 * YAC_RAD; buffer_lat[1] = -0.5 * YAC_RAD;

    yac_cdef_points_unstruct ( grid_ids[igrid],
                               nbr_cells,
                               YAC_LOCATION_CELL,
                               buffer_lon,
                               buffer_lat,
                               &cell_point_ids[igrid] );

    free ( buffer_lon );
    free ( buffer_lat );

    // Mask generation

    cell_mask = xmalloc (nbr_cells * sizeof(*cell_mask));

    for ( i = 0; i < nbr_cells; ++i )
      cell_mask[i] = 1;

    yac_cset_mask ( cell_mask,
                    cell_point_ids[igrid] );

    free (cell_mask);
  }

  field_id = xmalloc ( no_of_fields * sizeof(*field_id));

  // atmosphere output

  for ( i = 0; i < no_of_fields-4; i++ )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     &cell_point_ids[1],
                     1,
                     &field_id[i] );

  // ocean output

  for ( i = no_of_fields-4; i < no_of_fields; i++ )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     &cell_point_ids[0],
                     1,
                     &field_id[i] );

  yac_csearch ( &ierror );

  yac_cfinalize ();

  MPI_Finalize ();

  return 0;
}

