/**
 * 
 */
package de.mpg.mpimet.yacgui.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.mpg.mpimet.yacgui.controller.GuiController;
import de.mpg.mpimet.yacgui.model.Model.Align;
import de.mpg.mpimet.yacgui.ui.TransientPanel.TransientPanelData;
import de.mpg.mpimet.yacgui.ui.TransientPanel.onMouseCursorViewEnterEventListener;
import de.mpg.mpimet.yacgui.utils.Utils;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Panel that contains component's data and control buttons above 
 * 
 * @author Maxim Yastremsky
 */
public class ComponentPanel extends JPanel implements ChangeListener, onMouseCursorViewEnterEventListener {

	private static final long serialVersionUID = 616085318672002223L;

	private static final String DEFAULT_TITLE_LEFT = "Component 1";
	private static final String DEFAULT_TITLE_RIGHT = "Component 2";
	private static final String BUTTON_TOOLTIP = "Click to load component";
	private static final String BODY_TITLE = "Transients";
	private static final int BODY_GAP = 10;

	private JPanel headerPanel;
	private JPanel bodyPanel;
	private JLabel titleLabel;
	
	private JButton openButton;
	private JButton closeButton;
	
	private HashMap<Integer, TransientPanel> transientPanels;

	private Align align;
	private GuiController guiController;

	private JViewport viewPort;

	public ComponentPanel(Align align) {
		this.align = align;
		setLayout(new BorderLayout());
		guiController = GuiController.getInstance();
		transientPanels = new HashMap<Integer, TransientPanel>();
		initHeader();
		initBody();
	}

	public void setComponentPanelTitle(String title) {
		if (title == null || title.equals("")) {
			title = align.isLeft() ? DEFAULT_TITLE_LEFT : DEFAULT_TITLE_RIGHT;
		}
		
		titleLabel.setText(title);
	}
	
	public boolean isEmpty() {
		return transientPanels.isEmpty();
	}

	/**
	 * Loads transient panels data to component panel
	 * 
	 * @param panelDataList - list of panels data
	 */
	public void loadTransientPanelsData(List<TransientPanelData> panelDataList) {
		transientPanels.clear();
		
		for (int i = 0; i < panelDataList.size(); i++) {
			TransientPanel panel = new TransientPanel(this, align, i, panelDataList.get(i));
			transientPanels.put(i, panel);
		}
		closeButton.setEnabled(!transientPanels.isEmpty());
	}
	
	public void setAllCheckBoxesEnabled(boolean enabled) {
		for (TransientPanel panel : transientPanels.values()) {
			panel.setCheckBoxEnabled(enabled);
		}
	}
	
	public boolean areAllCheckBoxesEnabled() {
		for (TransientPanel panel : transientPanels.values()) {
			if (!panel.isCheckBoxEnabled()) {
				return false;
			}
		}
		return true;
	}
	
	public boolean areAllCheckBoxesDisabled() {
		for (TransientPanel panel : transientPanels.values()) {
			if (panel.isCheckBoxEnabled()) {
				return false;
			}
		}
		return true;
	}
	
	public void setCheckBoxEnabled(int index, boolean enabled) {
		transientPanels.get(index).setCheckBoxEnabled(enabled);
	}
	
	public void uncheckAllCheckBoxes() {
		for (TransientPanel panel : transientPanels.values()) {
			panel.setChecked(false);
		}
	}
	
	public void setCheckBoxChecked(int index, boolean checked) {
		transientPanels.get(index).setChecked(checked);
	}
	
	public TransientPanel getTransientPanel(int index) {
		return transientPanels.get(index);
	}
	
	public Point getRelativeCheckBoxArrowPosition(int index, ConnectionsGlassPane glassPane) {
		Point arrowPos = transientPanels.get(index).getArrowPosition();
		return SwingUtilities.convertPoint(transientPanels.get(index), arrowPos, glassPane);
	}
	
	@Override
	public void paintComponent(Graphics graphics) {
		super.paintComponent(graphics);
		this.setBackground(Color.white);
		graphics.setColor(Color.black);

		bodyPanel.removeAll();

		if (!transientPanels.isEmpty()) {
			GridBagConstraints c = new GridBagConstraints();
			for (int i = 0; i < transientPanels.size(); i++) {
				
				// add TransientPanel to body
				c.fill = GridBagConstraints.HORIZONTAL;
				c.gridx = 0;
				c.gridy = i*2;
				c.weightx = 1;
				bodyPanel.add(transientPanels.get(i), c);
				
				// leave some space
				c.gridx = 0;
				c.gridy = i * 2 + 1;
				bodyPanel.add(Box.createRigidArea(new Dimension(0, BODY_GAP)), c);
			}
			
			// fill space on bottom to make transients start on top
			c.weighty = 1;
			bodyPanel.add(Box.createRigidArea(new Dimension(0, BODY_GAP)), c);
		}

		validate();
	}

	private void initHeader() {
		headerPanel = new JPanel(new BorderLayout());
		
		ImageIcon openIcon = Utils.createImageIcon("/img/open-file-icon.png");
		openButton = new JButton(openIcon);
		headerPanel.add(openButton, BorderLayout.LINE_START);
		openButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				guiController.openComponent(align, null);
			}
		});
		
		titleLabel = new JLabel(align.isLeft() ? DEFAULT_TITLE_LEFT : DEFAULT_TITLE_RIGHT);
		titleLabel.setToolTipText(BUTTON_TOOLTIP);
		titleLabel.setBorder(new EmptyBorder(5, 5, 5, 5));
		headerPanel.add(titleLabel, BorderLayout.CENTER);
		
		ImageIcon closeIcon = Utils.createImageIcon("/img/close-icon.png");
		closeButton = new JButton(closeIcon);
		closeButton.setEnabled(false);
		closeButton.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				guiController.removeComponent(align);
				
			}
		});
		headerPanel.add(closeButton, BorderLayout.LINE_END);
		
		add(headerPanel, BorderLayout.NORTH);
	}

	private void initBody() {
		bodyPanel = new JPanel();
		bodyPanel.setLayout(new GridBagLayout());

		JScrollPane listScroller = new JScrollPane(bodyPanel);
		listScroller.setBorder(BorderFactory.createTitledBorder(BODY_TITLE));
		viewPort = listScroller.getViewport();
		viewPort.addChangeListener(this);
		
		add(listScroller, BorderLayout.CENTER);
	}

	@Override
	public void stateChanged(ChangeEvent e) {
		guiController.repaintGlassPane();
	}
	
	public boolean isArrowPointVisible(int index) {
		if (transientPanels.isEmpty()) {
			return false;
		}
		
		TransientPanel chosenTransientPanel = transientPanels.get(index); 
		Point arrowPointPos = chosenTransientPanel.getArrowPosition();
		
		//get point of arrow position in the main ComponentPanel
		Point arrowPosInPanel = SwingUtilities.convertPoint(chosenTransientPanel, arrowPointPos, bodyPanel);
		
		Rectangle visibleRect = viewPort.getVisibleRect();
		visibleRect.setLocation(viewPort.getViewPosition());
		
		return visibleRect.contains(arrowPosInPanel);
	}

	@Override
	public void onMouseCursorViewExit(TransientPanel cursorEnteredPanel) {
		for (TransientPanel panel : transientPanels.values()) {
			if (panel != cursorEnteredPanel) {
				panel.highlightBackground(false);
			}
		}
	}
}
