/**
 * @file test_config_xml.c
 *
 * @copyright Copyright  (C)  2014 DKRZ, MPI-M
 *
 * @author Teresa Holfeld <teresa.holfeld@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "tests.h"
#include "utils.h"
#include "xml/config_xml.h"

/**
 * Executes all given interface functions. Based on examples/config_xml_example.c
 * written by Teresa Holfeld.
 *
 */

// macros for indices
#define FIRST          0
#define SECOND         1
#define THIRD          2
#define FOURTH         3
#define THIRTEENTH    12

// macros for IDs
#define FIRST_ID       1
#define SECOND_ID      2
#define THIRD_ID       3
#define FOURTH_ID      4
#define SIXTH_ID       6
#define THIRTEENTH_ID 13

// constants for component names
static const char FIRST_NAME[] = "ICON-ocean";
//static const char EMPTY[] = "";   // enable for testing for errors
static const char NON_EXISTENT[] = "Non-existent";
static const char TRANSIENT_STANDARD_NAME_EXAMPLE[] = "sea_surface_temperature";

int main(void) {
   //* initialization *********************************************************/

   char * xml_filename = "coupling.xml"; // file ok
   char * schema_filename = "coupling.xsd";
   struct config_xml * config =
      yac_config_xml_new(xml_filename, schema_filename);

   int exists;

   //* components *************************************************************/

#ifdef VERBOSE
   printf("\nTesting components:\n\n");
#endif

   int a = yac_xml_get_number_of_components(config);
#ifdef VERBOSE
   printf("Number of components: %d\n", a);
#endif
   if ( a != 3 ) PUT_ERR ("ERROR: unexpected number of components.\n");
   // expected: exists == 1
   exists = yac_xml_component_name_exists_in_xml(config, FIRST_NAME);

#ifdef VERBOSE
   if (exists) {
      printf("Component with name '%s' exists. \n", FIRST_NAME);
   }
#endif
   if (!exists) PUT_ERR ("ERROR: ERROR: unexpected component name.\n");

   // expected: exists == 0
   exists = yac_xml_component_name_exists_in_xml(config, NON_EXISTENT);
#ifdef VERBOSE
   if (!exists) {
      printf("Component with name '%s' does not exist. \n", NON_EXISTENT);
   }
#endif
   if (exists) PUT_ERR ("ERROR: ERROR: unexpected component name.\n");

   // expected: exists == 1
   exists = yac_xml_component_id_exists_in_xml(config, FIRST_ID);
#ifdef VERBOSE
   if (exists) {
      printf("Component with ID '%d' exists. \n", FIRST_ID);
   }
#endif
   if (!exists) PUT_ERR ("ERROR: ERROR: unexpected component id.\n");

   // expected: exists == 0
   exists = yac_xml_component_id_exists_in_xml(config, THIRTEENTH_ID);
#ifdef VERBOSE
   if (!exists) {
      printf("Component with ID '%d' does not exist. \n", THIRTEENTH_ID);
   }
#endif
   if (exists) PUT_ERR ("ERROR: ERROR: unexpected component name.\n");

   int b = yac_xml_get_component_id(config, FIRST);
#ifdef VERBOSE
   printf("ID of component %d: %d\n", FIRST, b);
#endif
   if ( b != 1 ) PUT_ERR ("ERROR: ERROR: unexpected component id.\n");

   int c = yac_xml_get_component_id_by_component_name(config, FIRST_NAME);
#ifdef VERBOSE
   printf("ID of component '%s': %d \n", FIRST_NAME, c);
#endif
   if ( c != 1 ) PUT_ERR ("ERROR: ERROR: unexpected component id.\n");

   yac_xml_component * a_component = xmalloc(sizeof(yac_xml_component));
   yac_xml_get_component(config, FIRST, a_component);
#ifdef VERBOSE
   printf("Get component %d: name %s and id %d\n", FIRST, a_component->name,
         a_component->global_id);
#endif
   yac_xml_free_xml_component(a_component);

   yac_xml_component * b_component = xmalloc(sizeof(yac_xml_component));
   yac_xml_get_component_by_component_id(config, SECOND_ID, b_component);
#ifdef VERBOSE
   printf("Get component by ID %d: name %s and id %d\n", SECOND_ID,
         b_component->name, b_component->global_id);
#endif
   yac_xml_free_xml_component(b_component);

   yac_xml_component * c_component = xmalloc(sizeof(yac_xml_component));
   yac_xml_get_component_by_component_name(config, FIRST_NAME, c_component);
#ifdef VERBOSE
   printf("Component by name w/ name %s and id %d\n", c_component->name,
         c_component->global_id);
#endif
   yac_xml_free_xml_component(c_component);

   const char * a_name = yac_xml_get_component_name(config, FIRST);
#ifdef VERBOSE
   printf("Get component name of %d: %s \n", FIRST, a_name);
#endif
   if (strcmp(a_name, "ICON-ocean") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component name.\n");

   const char * b_name =
      yac_xml_get_component_name_by_component_id(config, FIRST_ID);
#ifdef VERBOSE
   printf("Get component name by component ID %d: %s \n", FIRST_ID, b_name);
#endif
   if (strcmp(b_name, "ICON-ocean") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component name.\n");

   const char * a_model = yac_xml_get_component_model(config, FIRST);
#ifdef VERBOSE
   printf("Get component model of %d: %s \n", FIRST, a_model);
#endif
   if (strcmp(a_model, "some_model...") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component model.\n");

   const char * b_model =
      yac_xml_get_component_model_by_component_id(config, FIRST_ID);
#ifdef VERBOSE
   printf("Get component model by component ID %d: %s \n", FIRST_ID, b_model);
#endif
   if (strcmp(b_model, "some_model...") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component model.\n");

   const char * a_simulated =
      yac_xml_get_component_simulated(config, FIRST);
#ifdef VERBOSE
   printf("Get component simulated of %d: %s \n", FIRST, a_simulated);
#endif
   if (strcmp(a_simulated, "ocean") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component.\n");

   const char * b_simulated =
      yac_xml_get_component_simulated_by_component_id(config, FIRST_ID);
#ifdef VERBOSE
   printf("Get component simulated by ID %d: %s \n", FIRST_ID, b_simulated);
#endif
   if (strcmp(b_simulated, "ocean") == 1)
      PUT_ERR ("ERROR: ERROR: unexpected component.\n");

   int a_no_t =
      yac_xml_get_number_of_transient_grid_refs(config, FIRST);
#ifdef VERBOSE
   printf("Get number of transients in component %d: %d \n", FIRST, a_no_t);
#endif
   if ( a_no_t != 5 )
      PUT_ERR ("ERROR: ERROR: unexpected number of transients.\n");

   int b_no_t =
      yac_xml_get_number_of_transient_grid_refs_by_component_id(
         config, FIRST_ID);
#ifdef VERBOSE
   printf("Get number of transients in component by component ID %d: %d \n",
         FIRST_ID, b_no_t);
#endif
   if ( b_no_t != 5 )
      PUT_ERR ("ERROR: ERROR: unexpected number of transients.\n");

   int transient_grid_ref_id =
      yac_xml_get_transient_grid_ref_id(config, FIRST, THIRD);
#ifdef VERBOSE
   printf(
      "Get transient grid ref id of transient_grid_ref no. %d of component no. "
      "%d: %d \n", THIRD, FIRST, transient_grid_ref_id);
#endif
   if ( transient_grid_ref_id != 3 )
      PUT_ERR ("ERROR: ERROR: unexpected transient_grid_ref_id.\n");

   int transient_grid_ref_id_by_component_id =
      yac_xml_get_transient_grid_ref_id_by_component_id(
         config, FIRST_ID, THIRD);
#ifdef VERBOSE
   printf(
      "Get transient grid ref id of transient_grid_ref no. %d of component "
      "with ID %d: %d \n",
      THIRD, FIRST_ID, transient_grid_ref_id_by_component_id);
#endif
   if ( transient_grid_ref_id_by_component_id != 3 )
	   PUT_ERR (
        "ERROR: ERROR: unexpected transient_grid_ref_id_by_component_id.\n");

   int transient_ref =
      yac_xml_get_transient_id_by_transient_grid_ref_id(
         config, FIRST, SECOND_ID);
#ifdef VERBOSE
   printf(
      "Get transient id by transient_grid_ref with ID %d in component with "
      "index %d: %d \n", SECOND_ID, FIRST, transient_ref);
#endif
   if ( transient_ref != 1 )
      PUT_ERR ("ERROR: ERROR: unexpected transient_ref.\n");

   int transient_ref_by_component_id =
      yac_xml_get_transient_id_by_transient_grid_ref_id_by_component_id(
         config, FIRST_ID, SECOND_ID);
#ifdef VERBOSE
   printf(
      "Get transient id by transient_grid_ref with ID %d in component with ID "
      "%d: %d \n", SECOND_ID, FIRST_ID, transient_ref_by_component_id);
#endif
   if ( transient_ref_by_component_id != 1 )
      PUT_ERR ("ERROR: ERROR: unexpected transient_ref_by_component_id.\n");

   int grid_ref =
      yac_xml_get_grid_id_by_transient_grid_ref_id(
         config, SECOND, FIRST_ID);
#ifdef VERBOSE
   printf(
      "Get grid id by transient_grid_ref with ID %d in component with index "
      "%d: %d \n", FIRST_ID, SECOND, grid_ref);
#endif
   if ( grid_ref != 2 ) PUT_ERR ("ERROR: ERROR: unexpected grid_ref.\n");

   int grid_ref_by_component_id =
      yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
         config, SECOND_ID, FIRST_ID);
#ifdef VERBOSE
   printf(
      "Get grid id by transient_grid_ref with ID %d in component with ID %d: "
      "%d \n", FIRST_ID, SECOND_ID, grid_ref_by_component_id);
#endif
   if ( grid_ref_by_component_id != 2 )
      PUT_ERR ("ERROR: ERROR: unexpected grid_ref_by_component_id.\n");

   int collection_size =
      yac_xml_get_transient_collection_size(config, SECOND_ID, FIRST_ID);

#ifdef VERBOSE
   printf(
      "Get collection size by transient_grid_ref with ID %d in component with "
      "ID %d: %d \n", FIRST_ID, SECOND_ID, collection_size);
#endif
   if ( collection_size != 3 )
      PUT_ERR ("ERROR: ERROR: unexpected collection_size.\n");

   //* transients *************************************************************/

#ifdef VERBOSE
   printf("\nTesting transients:\n\n");
#endif

   int no_t_total = yac_xml_get_number_of_transients(config);
#ifdef VERBOSE
   printf("Number of transients: %d \n", no_t_total);
#endif
   if ( no_t_total != 9 ) PUT_ERR ("ERROR: unexpected no_t_total.\n");

   int t_id =
      yac_xml_get_transient_id_by_transient_standard_name(
         config, TRANSIENT_STANDARD_NAME_EXAMPLE);
#ifdef VERBOSE
   printf("Get transient ID by transient standard name '%s': %d \n",
         TRANSIENT_STANDARD_NAME_EXAMPLE, t_id);
#endif
   if ( t_id != 1 ) PUT_ERR ("ERROR: unexpected transient_id.\n");

  // expected: exists == 1
   exists = yac_xml_transient_id_exists_in_xml(config, FIRST_ID);
#ifdef VERBOSE
   if (exists) {
      printf("Transient with ID '%d' exists. \n", FIRST_ID);
   }
#endif
   if ( !exists ) PUT_ERR ("ERROR: nonexisting entry.\n");

   // expected: exists == 0
   exists = yac_xml_transient_id_exists_in_xml(config, THIRTEENTH_ID);
#ifdef VERBOSE
   if (!exists) {
      printf("Transient with ID '%d' does not exist. \n", THIRTEENTH_ID);
   }
#endif
   if ( exists ) PUT_ERR ("ERROR: unexpected existing entry.\n");

   const char * a_standard_name =
      yac_xml_get_transient_standard_name_by_transient_id(
         config, FIRST_ID);
#ifdef VERBOSE
   printf("Get transient standard name by transient ID %d: %s \n", FIRST_ID,
         a_standard_name);
#endif
   yac_xml_free_xml_string(a_standard_name);

   // expected: exists == 1
   exists =
      yac_xml_transient_standard_name_exists_in_xml(
         config, TRANSIENT_STANDARD_NAME_EXAMPLE);
#ifdef VERBOSE
   if (exists) {
      printf("Transient with standard name '%s' exists. \n",
            TRANSIENT_STANDARD_NAME_EXAMPLE);
   }
#endif
   if ( !exists ) PUT_ERR ("ERROR: nonexisting entry.\n");

   // expected: exists == 0
   exists =
      yac_xml_transient_standard_name_exists_in_xml(config, NON_EXISTENT);
#ifdef VERBOSE
   if (!exists) {
      printf("Transient with standard name '%s' does not exist. \n",
            NON_EXISTENT);
   }
#endif
   if ( exists ) PUT_ERR ("ERROR: unexpected entry.\n");

   yac_xml_transient * a_t = xmalloc(sizeof(yac_xml_transient));
   yac_xml_get_transient(config, FIRST, a_t);
#ifdef VERBOSE
   printf("Get transient %d: name '%s', ID %d \n", FIRST, a_t->name,
         a_t->global_id);
#endif
   yac_xml_free_xml_transient(a_t);

   yac_xml_transient * t_by_id = xmalloc(sizeof(yac_xml_transient));
   yac_xml_get_transient_by_id(config, FIRST_ID, t_by_id);
#ifdef VERBOSE
   printf("Get transient by ID %d: name '%s', ID %d \n", FIRST_ID,
         t_by_id->name, t_by_id->global_id);
#endif
   yac_xml_free_xml_transient(t_by_id);

   yac_xml_transient * transients =
      xmalloc(no_t_total * sizeof(yac_xml_transient));
   yac_xml_get_all_transients(config, transients);
#ifdef VERBOSE
   printf("Get all transients: \n");
   for (int i = 0; i < no_t_total; i++) {
      printf("   Standard name: '%s', ID: %d \n", transients[i].name,
            transients[i].global_id);
   }
#endif
   yac_xml_free_xml_transients(transients, no_t_total);

   //* grids ******************************************************************/

#ifdef VERBOSE
   printf("\nTesting grids:\n\n");
#endif

   int a_no_grids = yac_xml_get_number_of_grids(config);
#ifdef VERBOSE
   printf("Number of grids: %d \n", a_no_grids);
#endif

   // expected: exists == 1
   exists = yac_xml_grid_id_exists_in_xml(config, FIRST_ID);
#ifdef VERBOSE
   if (exists) {
      printf("Grid with ID '%d' exists. \n", FIRST_ID);
   }
#endif
   if ( !exists ) PUT_ERR ("ERROR: nonexisting entry.\n");

   // expected: exists == 0
   exists = yac_xml_grid_id_exists_in_xml(config, THIRTEENTH_ID);
#ifdef VERBOSE
   if (!exists) {
      printf("Grid with ID '%d' does not exist. \n", THIRTEENTH_ID);
   }
#endif
   if ( exists ) PUT_ERR ("ERROR: unexpected entry.\n");

   const char * a_alias_name =
      yac_xml_get_grid_alias_name_by_grid_id(config, FIRST_ID);
#ifdef VERBOSE
   printf("Get grid alias name by grid ID %d: %s \n", FIRST_ID, a_alias_name);
#endif
   yac_xml_free_xml_string(a_alias_name);

   // expected error: no such ID
   //   const char * a_error_name = get_grid_alias_name_by_grid_id(THIRTEENTH_ID);

   yac_xml_grid * g = xmalloc(sizeof(yac_xml_grid));
   yac_xml_get_grid(config, FIRST, g);
#ifdef VERBOSE
   printf("Get grid %d: ID %d, alias name '%s' \n", FIRST, g->global_id,
         g->name);
#endif
   yac_xml_free_xml_grid(g);

   yac_xml_grid * g_by_id = xmalloc(sizeof(yac_xml_grid));
   yac_xml_get_grid_by_id(config, SECOND_ID, g_by_id);
#ifdef VERBOSE
   printf("Get grid by ID %d: ID %d, alias name '%s' \n", SECOND_ID,
         g_by_id->global_id, g_by_id->name);
#endif
   yac_xml_free_xml_grid(g_by_id);

   yac_xml_grid * grids = xmalloc(a_no_grids * sizeof(yac_xml_grid));
   yac_xml_get_all_grids(config, grids);
#ifdef VERBOSE
   for (int i = 0; i < a_no_grids; i++) {
      printf("   Alias name: '%s', ID: %d \n", grids[i].name,
            grids[i].global_id);
   }
#endif
   yac_xml_free_xml_grids(grids, a_no_grids);

   //* dates and timestep *****************************************************/

#ifdef VERBOSE
   printf("\nTesting dates and timestep:\n\n");
#endif

   const char * start_date = yac_xml_get_start_date(config);
#ifdef VERBOSE
   printf("Start date: %s \n", start_date);
#endif
   if (strcmp(start_date, "2008-03-09T16:05:07") == 1)
      PUT_ERR ("ERROR: wrong start date.\n");

   const char * end_date = yac_xml_get_end_date(config);
#ifdef VERBOSE
   printf("End date: %s \n", end_date);
#endif
   if (strcmp(end_date, "2008-03-10T16:05:07") == 1)
      PUT_ERR ("ERROR: wrong end date.\n");

   yac_xml_timestep_unit unit = yac_xml_get_timestep_unit(config);
#ifdef VERBOSE
   switch (unit) {
   case T_SECOND:
      printf("Timestep unit: T_SECOND \n");
      break;
   case T_MINUTE:
      printf("Timestep unit: T_MINUTE \n");
      break;
   case T_HOUR:
      printf("Timestep unit: T_HOUR \n");
      break;
   case T_DAY:
      printf("Timestep unit: T_DAY \n");
      break;
   case T_MONTH:
      printf("Timestep unit: T_MONTH \n");
      break;
   case T_YEAR:
      printf("Timestep unit: T_YEAR \n");
      break;
   case T_MILLISECOND:
      printf("Timestep unit: T_MILLISECOND \n");
      break;
   default:
      printf("Error: timestep unit undefined. \n");
      break;
   }
#else
   switch (unit) {
   case T_SECOND:
      break;
   case T_MINUTE:
      break;
   case T_HOUR:
      break;
   case T_DAY:
      break;
   case T_MONTH:
      break;
   case T_YEAR:
      break;
   case T_MILLISECOND:
      break;
   default:
	  PUT_ERR ("ERROR: timestep unit undefined..\n");
      break;
   }
#endif

   //* couples ****************************************************************/

#ifdef VERBOSE
   printf("\nTesting couples:\n\n");
#endif

   int couples = yac_xml_get_number_of_couples(config);
#ifdef VERBOSE
   printf("Number of couples: %d\n", couples);
#endif
   if ( couples != 1 ) PUT_ERR ("ERROR: unexpected number of couples.\n");

   int first_component_id = yac_xml_get_first_component_id(config, FIRST);
#ifdef VERBOSE
   printf("First component ID of couple %d: %d \n", FIRST, first_component_id);
#endif
   if ( first_component_id != 1 )
      PUT_ERR ("ERROR: unexpected first_component_id.\n");

   int second_component_id = yac_xml_get_second_component_id(config, FIRST);
#ifdef VERBOSE
   printf("Second component ID of couple %d: %d \n", FIRST,
         second_component_id);
#endif
   if ( second_component_id != 2 )
      PUT_ERR ("ERROR: unexpected second_component_id.\n");

   int no_t_c = yac_xml_get_number_of_transient_couples(config, FIRST);
#ifdef VERBOSE
   printf("Number of transient couples of couple %d: %d \n", FIRST, no_t_c);
#endif
   if ( no_t_c != 2 )
      PUT_ERR ("ERROR: unexpected number of transient couples.\n");

   int t_c_id = yac_xml_get_transient_couple_transient_id(config, FIRST, FIRST);
#ifdef VERBOSE
   printf(
      "Get transient couple transient ID (of couple %d, transient couple %d): "
      "%d \n", FIRST, FIRST, t_c_id);
#endif
   if ( t_c_id != 1 ) PUT_ERR ("ERROR: unexpected transient ID.\n");

   yac_xml_transient * t_c_t = xmalloc(sizeof(yac_xml_transient));
   yac_xml_get_transient_couple_transient(config, FIRST, FIRST, t_c_t);
#ifdef VERBOSE
   printf(
      "Get transient couple transient (od couple %d, transient %d): standard "
      "name '%s', ID %d \n", FIRST, FIRST, t_c_t->name, t_c_t->global_id);
#endif
   if ( t_c_t->global_id != 1 ) PUT_ERR ("ERROR: unexpected global ID.\n");
   yac_xml_free_xml_transient(t_c_t);

   int k = yac_xml_get_source_component_id(config, FIRST, SECOND);
#ifdef VERBOSE
   printf("Get source component ID (of couple %d, transient couple %d): %d \n",
         FIRST, SECOND, k);
#endif
   if ( k != 2 ) PUT_ERR ("ERROR: unexpected ID.\n");

   int l = yac_xml_get_target_component_id(config, FIRST, SECOND);
#ifdef VERBOSE
   printf("Get target component ID (of couple %d, transient couple %d): %d \n",
         FIRST, SECOND, l);
#endif
   if ( l != 1 ) PUT_ERR ("ERROR: unexpected ID.\n");

   int source_transient_grid_ref_id =
      yac_xml_get_source_transient_grid_ref_id(
         config, FIRST, SECOND);
#ifdef VERBOSE
   printf(
      "Get source transient_grid_ref ID (of couple %d, transient couple %d): "
      "%d \n", SECOND, FIRST, source_transient_grid_ref_id);
#endif
   if ( source_transient_grid_ref_id != 3 ) PUT_ERR ("ERROR: unexpected ID.\n");

   int target_transient_grid_ref_id =
      yac_xml_get_target_transient_grid_ref_id(
         config, FIRST, SECOND);
#ifdef VERBOSE
   printf(
      "Get target transient_grid_ref ID (of couple %d, transient couple %d): "
      "%d \n", SECOND, FIRST, target_transient_grid_ref_id);
#endif
   if ( target_transient_grid_ref_id != 3 ) PUT_ERR ("ERROR: unexpected ID.\n");

   int source_timestep = yac_xml_get_source_timestep(config, FIRST, FIRST);
#ifdef VERBOSE
   printf("Get source timestep (of couple %d, transient couple %d): %d \n",
         FIRST, FIRST, source_timestep);
#endif
   if ( source_timestep != 10 ) PUT_ERR ("ERROR: unexpected timestep.\n");

   int target_timestep = yac_xml_get_target_timestep(config, FIRST, FIRST);
#ifdef VERBOSE
   printf("Get target timestep (of couple %d, transient couple %d): %d \n",
            FIRST, FIRST, target_timestep);
#endif
   if ( target_timestep != 20 ) PUT_ERR ("ERROR: unexpected timestep.\n");

   int coupling_period = yac_xml_get_coupling_period(config, FIRST, SECOND);
#ifdef VERBOSE
   printf("Get coupling period (of couple %d, transient couple %d): %d \n",
         FIRST, SECOND, coupling_period);
#endif
   if ( coupling_period != 60 )
      PUT_ERR ("ERROR: unexpected coupling_period.\n");

   yac_xml_operation o =
      yac_xml_get_coupling_period_operation(
         config, FIRST, SECOND);
#ifdef VERBOSE
   if (o == T_ACCUMULATE) {
      printf(
         "Get coupling period operation (of couple %d, transient couple %d): "
         "accumulate \n", FIRST, SECOND);
   } else if (o == T_AVERAGE) {
      printf(
         "Get coupling period operation (of couple %d, transient couple %d): "
         "average \n", FIRST, SECOND);
   } else if (o == T_NONE) {
      printf(
         "Get coupling period operation (of couple %d, transient couple %d): "
         "none \n", FIRST, SECOND);
   }
#endif
   if (o != T_ACCUMULATE) PUT_ERR ("ERROR: unexpected time operation.\n");

   int sourcelag = yac_xml_get_source_timelag(config, FIRST, SECOND);
#ifdef VERBOSE
   printf(
      "Get source timelag (of couple %d, transient couple %d): %d \n",
      FIRST, SECOND, sourcelag);
#endif
   if ( sourcelag != 0 ) PUT_ERR ("ERROR: wrong ID.\n");

   int targetlag = yac_xml_get_target_timelag(config, FIRST, SECOND);
#ifdef VERBOSE
   printf(
      "Get target timelag (of couple %d, transient couple %d): %d \n",
      FIRST, SECOND, targetlag);
#endif
   if ( targetlag != 0 ) PUT_ERR ("ERROR: wrong ID.\n");

   int no_interpolations =
      yac_xml_get_number_of_interpolations(config, FIRST, SECOND);
#ifdef VERBOSE
   printf("Number of interpolations (of couple %d, transient couple %d): %d \n",
         FIRST, SECOND, no_interpolations);
#endif
   if ( no_interpolations != 3 )
      PUT_ERR ("ERROR: wrong number of interpolations.\n");

   int equals =
      yac_xml_compare_interpolation_lists(
         config, FIRST, FIRST, FIRST, FIRST);
#ifdef VERBOSE
   printf("Interpolation lists are equal (0:0, 0:0): %d \n", equals);
#endif
   if ( !equals ) PUT_ERR ("ERROR: lists do not compare.\n");

   int equals_not =
      yac_xml_compare_interpolation_lists(config, FIRST, FIRST, FIRST, SECOND);
#ifdef VERBOSE
   printf("Interpolation lists are equal (0:0, 0:1): %d \n", equals_not);

   printf("Interpolations of couple %d, transient couple %d: \n", FIRST,
         SECOND);
#endif
   if ( equals_not ) PUT_ERR ("ERROR: lists should not be equal.\n");

   for (int i = 0; i < no_interpolations; i++) {
      enum yac_interpolation_list interpolation =
         yac_xml_get_interpolation_method(
            config, FIRST, SECOND, i);
      char * interpolation_name =
         yac_interpolation_enum_to_string(interpolation);
#ifdef VERBOSE
      printf("   Interpolation #%d: %s \n", i, interpolation_name);
#endif
      if ( i == 0 && strcmp(interpolation_name, "average") )
         PUT_ERR ("ERROR: yac_xml_get_interpolation_method average\n");
      if ( i == 1 && strcmp(interpolation_name, "n-nearest_neighbor") )
         PUT_ERR (
            "ERROR: yac_xml_get_interpolation_method n-nearest_neighbor\n");
      if ( i == 2 && strcmp(interpolation_name, "conservative") )
         PUT_ERR ("ERROR: yac_xml_get_interpolation_method conservative\n");
   }
   int has_parameters =
      yac_xml_interpolation_has_parameters(config, FIRST, SECOND, THIRD);
#ifdef VERBOSE
   if (has_parameters) {
      printf(
         "Interpolation %d in transient couple %d in couple %d has "
         "parameters. \n", THIRD, SECOND, FIRST);
   }
#endif
   if (!has_parameters)
      PUT_ERR(
         "ERROR: yac_xml_interpolation_has_parameters "
         "should provide parameters\n");

   has_parameters =
      yac_xml_interpolation_has_parameters(config, FIRST, FIRST, FIRST);
#ifdef VERBOSE
   if (!has_parameters) {
      printf(
         "Interpolation %d in transient couple %d in couple %d does not have "
         "parameters. \n", FIRST, FIRST, FIRST);
   }
#endif
   if (has_parameters)
      PUT_ERR(
         "ERROR: yac_xml_interpolation_has_parameters "
         "should not provide parameters\n");

   char * parameter = "n";
   int has_parameter =
      yac_xml_interpolation_has_parameter(
         config, FIRST, SECOND, SECOND, parameter);
#ifdef VERBOSE
   if (has_parameter) {
      printf("Interpolation %d in transient couple %d in couple %d has "
             "parameter '%s'. \n", SECOND, SECOND, FIRST, parameter);
   }
#endif
   if (!has_parameter)
      PUT_ERR(
         "ERROR: yac_xml_interpolation_has_parameters "
         "should provide parameters\n");

   has_parameter =
      yac_xml_interpolation_has_parameter(
         config, FIRST, SECOND, SECOND, NON_EXISTENT);
#ifdef VERBOSE
   if (!has_parameter) {
      printf(
         "Interpolation %d in transient couple %d in couple %d does "
         "not have parameter '%s'. \n", SECOND, SECOND, FIRST, NON_EXISTENT);
   }
#endif
   if (has_parameter)
      PUT_ERR("ERROR: yac_xml_interpolation_has_parameters should "
              "not provide parameters\n");

   int number_parameters =
      yac_xml_get_interpolation_number_of_parameters(
         config, FIRST, SECOND, THIRD);
#ifdef VERBOSE
   printf(
      "Interpolation %d in transient couple %d in couple %d has %d "
      "parameters. \n", THIRD, SECOND, FIRST, number_parameters);
#endif
   if (number_parameters != 3)
      PUT_ERR("ERROR: wrong number of parameters.\n");

   const char * parameter_name =
      yac_xml_get_interpolation_parameter_name(
         config, FIRST, SECOND, THIRD, FIRST);
#ifdef VERBOSE
   printf(
      "Parameter name of parameter #%d (couple %d, transient couple %d, "
      "interpolation %d): '%s' \n",
      FIRST, FIRST, SECOND, THIRD, parameter_name);
#endif
   if ( strcmp(parameter_name, "enforced_conservation") )
      PUT_ERR("ERROR: yac_xml_get_interpolation_parameter_name: wrong name\n");

   const char * parameter_value =
      yac_xml_get_interpolation_parameter_value_string(
         config, FIRST, SECOND, SECOND, parameter);
#ifdef VERBOSE
   printf(
      "Parameter value of attribute '%s' (couple %d, transient couple %d, "
      "interpolation %d): '%s' \n",
      parameter, FIRST, SECOND, SECOND, parameter_value);
#endif
   if (strcmp(parameter_value,"2")) PUT_ERR ("ERROR: wrong parameter value\n");

   yac_xml_free_xml_string(parameter_value);

   const int parameter_value_int =
      yac_xml_get_interpolation_parameter_value_int(
         config, FIRST, SECOND, SECOND, parameter);
#ifdef VERBOSE
   printf(
      "Parameter value (int) of attribute '%s' (couple %d, transient couple "
      "%d, interpolation %d): %d \n",
      parameter, FIRST, SECOND, SECOND, parameter_value_int);
#endif
   if (parameter_value_int != 2)
      PUT_ERR ("ERROR: wrong int value for parameter.\n");

   const double parameter_value_double =
      yac_xml_get_interpolation_parameter_value_double(
         config, FIRST, SECOND, SECOND, parameter);
#ifdef VERBOSE
   printf(
      "Parameter value (double) of attribute '%s' (couple %d, transient couple "
      "%d, interpolation %d): %f \n",
      parameter, FIRST, SECOND, SECOND, parameter_value_double);
#endif
   if ((parameter_value_double > 2.0) !=
       (parameter_value_double < 2.0))
     PUT_ERR ("ERROR: wrong double value for parameter.\n");

   //* destroy ****************************************************************/

   yac_config_xml_delete(config);
   return TEST_EXIT_CODE;
}

