/**
 * @file test_cxc.c
 *
 *
 * @copyright Copyright  (C)  2014 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <math.h>

#include "utils.h"
#include "geometry.h"
#include "test_cxc.h"

static int test_points_vec(double p[3], double ref_p[3]) {

   double const tol = 1.0e-12;

   return get_vector_angle(p, ref_p) < tol;
}

static void test_results_vec(double p[3], double q[3], int ret_val,
                             double ref_p[3], double ref_q[3],
                             int ref_ret_value) {

   int num_matching_points = 0;
   int const mask_p = 1 + 4;
   int const mask_q = 2 + 8;
   int const shift_p_to_q = 1, shift_q_to_p = 1;
   int const identically_circles = 16;

   if (ref_ret_value == -1) {

      if (ret_val != -1) PUT_ERR("error in return value (does not indicate identically great circles)\n")

      return;
   }

   if ((ret_val & identically_circles) ^ (ref_ret_value & identically_circles))
      PUT_ERR("error in ret_val (identically circles)\n")

   // if there is only one intersection point
   if (test_points_vec(ref_q, ref_p)) {

      if (ret_val != ref_ret_value)
         PUT_ERR("error in ret_val (a)");

      if (test_points_vec(p, ref_p))
         ++num_matching_points;

      if (test_points_vec(q, ref_p))
         ++num_matching_points;

   } else {

      if (test_points_vec(p, ref_p)) {

         ++num_matching_points;
         if ((ret_val & mask_p)                 != (ref_ret_value & mask_p)) {
            PUT_ERR("error in ret_val (b)\n")
         }
      }
      if (test_points_vec(q, ref_p)) {

         ++num_matching_points;
         if ((ret_val & mask_q) >> shift_q_to_p != (ref_ret_value & mask_p)) {
            PUT_ERR("error in ret_val (c)\n")
         }
      }
      if (test_points_vec(p, ref_q)) {

         ++num_matching_points;
         if ((ret_val & mask_p) << shift_p_to_q != (ref_ret_value & mask_q)) {
            PUT_ERR("error in ret_val (d)\n")
         }
      }
      if (test_points_vec(q, ref_q)) {

         ++num_matching_points;
         if ((ret_val & mask_q)                 != (ref_ret_value & mask_q)) {
            PUT_ERR("error in ret_val (e)\n")
         }
      }
   }

   if (num_matching_points != 2) {
      PUT_ERR("p and/or q did not match\n")
   }

   if (ref_ret_value == 0 && ret_val != 0) PUT_ERR("error in return value (e)\n")
}

void test_cxc_rad(double lon_a, double lat_a, double lon_b, double lat_b,
                  double lon_c, double lat_c, double lon_d, double lat_d,
                  enum yac_edge_type edge_type_a, enum yac_edge_type edge_type_b,
                  double lon_ref_p, double lat_ref_p,
                  double lon_ref_q, double lat_ref_q, int ref_ret_val,
                  int (*intersect_func_vec)(double *, double *, double *,
                                            double *, double *, double *)) {

  double ref_p[3], ref_q[3];

  LLtoXYZ(lon_ref_p, lat_ref_p, ref_p);
  LLtoXYZ(lon_ref_q, lat_ref_q, ref_q);

  for (int i = 0; i < 4; ++i) {

    double edge_a[2][3], edge_b[2][3];

    LLtoXYZ(lon_a, lat_a, edge_a[i&1]);
    LLtoXYZ(lon_b, lat_b, edge_a[(i&1)^1]);
    LLtoXYZ(lon_c, lat_c, edge_b[i>>1]);
    LLtoXYZ(lon_d, lat_d, edge_b[(i>>1)^1]);

    double p[3], q[3];
    int ret_val;
    
    ret_val = intersect_func_vec(edge_a[0], edge_a[1],
                                 edge_b[0], edge_b[1], p, q);
    test_results_vec(p, q, ret_val, ref_p, ref_q, ref_ret_val);
  }
}

void test_cxc(double lon_a, double lat_a, double lon_b, double lat_b,
              double lon_c, double lat_c, double lon_d, double lat_d,
              enum yac_edge_type edge_type_a, enum yac_edge_type edge_type_b,
              double lon_ref_p, double lat_ref_p,
              double lon_ref_q, double lat_ref_q, int ref_ret_val,
              int (*intersect_func_vec)(double *, double *, double *,
                                        double *, double *, double *)) {

   test_cxc_rad(lon_a*YAC_RAD, lat_a*YAC_RAD, lon_b*YAC_RAD, lat_b*YAC_RAD,
                lon_c*YAC_RAD, lat_c*YAC_RAD, lon_d*YAC_RAD, lat_d*YAC_RAD,
                edge_type_a, edge_type_b, lon_ref_p*YAC_RAD, lat_ref_p*YAC_RAD,
                lon_ref_q*YAC_RAD, lat_ref_q*YAC_RAD, ref_ret_val,
                intersect_func_vec);
}

