/**
 * @file test_dummy_coupling2.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#define EXACT

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_interface.h"
#include "read_icon_grid.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"
#include "utils.h"
#include "tests.h"

#define DUMMY_VALUE (-1337.0)

struct field_config {
  double * data;
  size_t data_size;
  int id;
};

static void generate_icon_grid(
  int comm_rank, int comm_size, int comp_id, int * grid_id,
  struct field_config * in_config, struct field_config * out_config);
static void generate_cube_grid(
  int comm_rank, int comm_size, int comp_id, int * grid_id,
  struct field_config * in_config, struct field_config * out_config);
static void init_in_field_data(struct field_config field_config);
static void check_results(
  struct field_config field_config, struct field_config ref_field_config);

int main (void) {

  yac_cinit ("coupling_test2.xml", "coupling.xsd");
  int global_rank, global_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &global_size);

  if (global_size != 3) {
    fprintf(stderr, "Wrong number of processes (should be 3)\n");
    exit(EXIT_FAILURE);
  }

  // register component(s)
  // (rank 1 => comp1; rank 2 => comp2; rank 3 => comp1 + comp2)
  int comp_ids[2];
  int num_comps;
  if ((global_rank == 0) || (global_rank == 1)) {
    yac_cdef_comp((global_rank == 0)?"comp1":"comp2", comp_ids);
    num_comps = 1;
  } else {
    char const * comp_names[2] = {"comp1", "comp2"};
    yac_cdef_comps(&(comp_names[0]), 2, comp_ids);
    num_comps = 2;
  }

  // register grids
  int grid_ids[2];
  struct field_config in_field_configs[2];
  struct field_config out_field_configs[2];
  if ((global_rank == 0) || (global_rank == 2))
    generate_icon_grid(
      (global_rank == 2), 2, comp_ids[0], &(grid_ids[0]),
      &(in_field_configs[0]),
      &(out_field_configs[0]));

  if ((global_rank == 1) || (global_rank == 2))
    generate_cube_grid(
      (global_rank == 2), 2, comp_ids[global_rank == 2],
      &(grid_ids[global_rank == 2]),
      &(in_field_configs[global_rank == 2]),
      &(out_field_configs[global_rank == 2]));

  int ierror;
  yac_csearch(&ierror);

  // do some ping-pongs
  for (int t = 0; t < 100; ++t) {

    if (global_rank == 0) {

      {
        int info, err;
        int id = out_field_configs[0].id;
        double *point_set_data[1];
        double **collection_data[1] = {point_set_data};
        point_set_data[0] = out_field_configs[0].data;

        yac_cput(id, 1, collection_data, &info, &err);
      }

      {
        int info, err;
        int id = in_field_configs[0].id;
        double *collection_data[1] = {in_field_configs[0].data};

        init_in_field_data(in_field_configs[0]);

        yac_cget(id, 1, collection_data, &info, &err);

        check_results(in_field_configs[0], out_field_configs[0]);
      }
    } else if (global_rank == 1) {

      {
        int info, err;
        int id = in_field_configs[0].id;
        double *collection_data[1] = {in_field_configs[0].data};

        init_in_field_data(in_field_configs[0]);

        yac_cget(id, 1, collection_data, &info, &err);

        check_results(in_field_configs[0], out_field_configs[0]);
      }

      {
        int info, err;
        int id = out_field_configs[0].id;
        double *point_set_data[1];
        double **collection_data[1] = {point_set_data};
        point_set_data[0] = out_field_configs[0].data;

        yac_cput(id, 1, collection_data, &info, &err);
      }
    } else {

      for (int i = 0; i < 2; ++i) {
        int out_info, in_info, err;
        int out_id = out_field_configs[i].id;
        int in_id = in_field_configs[i^1].id;
        double *out_point_set_data[1];
        double **out_collection_data[1] = {out_point_set_data};
        out_point_set_data[0] = out_field_configs[i].data;
        double *in_collection_data[1] = {in_field_configs[i^1].data};

        init_in_field_data(in_field_configs[i^1]);

        yac_cexchange(
          out_id, in_id, 1, out_collection_data, in_collection_data,
          &out_info, &in_info, &err);

        check_results(in_field_configs[i^1], out_field_configs[i^1]);
      }
    }
  }

  // clean-up

  for (size_t i = 0; i < num_comps; ++i) {
    free(out_field_configs[i].data);
    free(in_field_configs[i].data);
  }

  yac_cfinalize();

  return TEST_EXIT_CODE;
}

static void generate_icon_grid(
  int comm_rank, int comm_size, int comp_id, int * grid_id,
  struct field_config * in_config, struct field_config * out_config) {

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * global_cell_id;
  int * cell_core_mask;
  int * global_corner_id;
  int * corner_core_mask;

  read_part_icon_grid_information(
    "icon_grid_R02B02.nc", &nbr_vertices, &nbr_cells, &num_vertices_per_cell,
    &cell_to_vertex, &x_vertices, &y_vertices, &x_cells, &y_cells,
    &global_cell_id, &cell_mask,
    &cell_core_mask, &global_corner_id, &corner_core_mask,
    comm_rank, comm_size);

  yac_cdef_grid_unstruct(
    "icon", nbr_vertices, nbr_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, grid_id);

  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, *grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, *grid_id);
  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, *grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, *grid_id);

  int corner_point_id, cell_point_id;
  yac_cdef_points_unstruct(
    *grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    &corner_point_id);
  yac_cdef_points_unstruct(
    *grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    &cell_point_id);

  for (size_t i = 0; i < nbr_cells; ++i) cell_mask[i] = cell_mask[i] == 0;

  yac_cset_mask(cell_mask, cell_point_id);

  yac_cdef_field(
    "icon_to_cube", comp_id, &cell_point_id, 1, &(out_config->id));
  yac_cdef_field(
    "cube_to_icon", comp_id, &cell_point_id, 1, &(in_config->id));

  out_config->data = xmalloc(nbr_cells * sizeof(*(out_config->data)));
  for (int i = 0; i < nbr_cells; ++i)
    out_config->data[i] =
      (cell_core_mask[i])?
        (test_harmonic(x_cells[i]*YAC_RAD, y_cells[i]*YAC_RAD)):DUMMY_VALUE;
  in_config->data = xmalloc(nbr_cells * sizeof(*(in_config->data)));

  in_config->data_size = nbr_cells;
  out_config->data_size = nbr_cells;

  delete_icon_grid_data(&cell_mask,
                        &global_cell_id,
                        &cell_core_mask,
                        &num_vertices_per_cell,
                        &global_corner_id,
                        &corner_core_mask,
                        &cell_to_vertex,
                        &x_cells,
                        &y_cells,
                        &x_vertices,
                        &y_vertices);
}

static void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** cell_core_mask,
  int ** global_corner_id, int ** corner_core_mask, int rank, int size) {

  double * x_vertices_3d, * y_vertices_3d, * z_vertices_3d;
  unsigned * dummy;

  // generate global grid
  generate_cubed_sphere_grid_information(n, nbr_cells, nbr_vertices,
                                         &x_vertices_3d, &y_vertices_3d,
                                         &z_vertices_3d, cell_to_vertex,
                                         &dummy);
  *num_vertices_per_cell = xmalloc(*nbr_cells * sizeof(**num_vertices_per_cell));
  for (unsigned i = 0; i < *nbr_cells; ++i) (*num_vertices_per_cell)[i] = 4;
  free(dummy);

  double * x_cell_3d = xmalloc(*nbr_cells * sizeof(*x_cell_3d));
  double * y_cell_3d = xmalloc(*nbr_cells * sizeof(*y_cell_3d));
  double * z_cell_3d = xmalloc(*nbr_cells * sizeof(*z_cell_3d));

  for (unsigned i = 0; i < *nbr_cells; ++i) {

    x_cell_3d[i] = y_cell_3d[i] = z_cell_3d[i] = 0;

    for (unsigned j = 0; j < 4; ++j) {

      x_cell_3d[i] += x_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      y_cell_3d[i] += y_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      z_cell_3d[i] += z_vertices_3d[(*cell_to_vertex)[4 * i + j]];
    }

    double scale = 1.0 / sqrt(x_cell_3d[i] * x_cell_3d[i] + 
                              y_cell_3d[i] * y_cell_3d[i] + 
                              z_cell_3d[i] * z_cell_3d[i]);

    x_cell_3d[i] *= scale;
    y_cell_3d[i] *= scale;
    z_cell_3d[i] *= scale;
  }

  // determine local rank
  int num_cells_per_process = (*nbr_cells + size - 1)/size;
  int local_start = rank * num_cells_per_process;
  int num_local_cells =
    MIN(num_cells_per_process, (int)(*nbr_cells) - local_start);

  // mask for required vertices and cells
  int * required_vertices = xcalloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = xcalloc(*nbr_cells, sizeof(*required_cells));

  int offset = 0;
  for (int i = 0; i < local_start; ++i)
    offset += (*num_vertices_per_cell)[i];

  // mark all local cells and their vertices as required
  for (int i = local_start; i < local_start + num_local_cells; ++i) {

    required_cells[i] = 1;
    for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo cells
  int * cell_is_on_rank = xmalloc(*nbr_cells * sizeof(*cell_is_on_rank));

  // mark all halo cells as required and generate cell_is_on_rank
  offset = 0;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (!required_cells[i]) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          cell_is_on_rank[i] = i / num_cells_per_process;
          break;
        }
      }

    } else {
      cell_is_on_rank[i] = -1;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo vertices
  int * vertex_is_on_rank = xmalloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  offset = 0;
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (unsigned i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = xmalloc(part_num_cells * sizeof(**global_cell_id));
  *cell_core_mask = xmalloc(part_num_cells * sizeof(**cell_core_mask));
  *global_corner_id = xmalloc(part_num_vertices * sizeof(**global_corner_id));
  *corner_core_mask = xmalloc(part_num_vertices * sizeof(**corner_core_mask));

  *x_cells = xmalloc(part_num_cells * sizeof(**x_cells));
  *y_cells = xmalloc(part_num_cells * sizeof(**y_cells));
  *x_vertices = xmalloc(part_num_vertices * sizeof(**x_vertices));
  *y_vertices = xmalloc(part_num_vertices * sizeof(**y_vertices));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = xmalloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (unsigned i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*corner_core_mask)[part_num_vertices] = vertex_is_on_rank[i] == -1;
      double p[3] = {x_vertices_3d[i], y_vertices_3d[i], z_vertices_3d[i]};
      XYZtoLL(p, (*x_vertices) + part_num_vertices, (*y_vertices) + part_num_vertices);
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  offset = 0;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*cell_core_mask)[part_num_cells] = (cell_is_on_rank[i] == -1);
      double middle_point[3] = {x_cell_3d[i],y_cell_3d[i],z_cell_3d[i]};

      

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {
        (*cell_to_vertex)[num_cell_vertex_dependencies+j] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];
      }

      XYZtoLL(middle_point, (*x_cells)+part_num_cells, (*y_cells)+part_num_cells);

      num_cell_vertex_dependencies += (*num_vertices_per_cell)[i];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  free(x_cell_3d);
  free(y_cell_3d);
  free(z_cell_3d);
  free(x_vertices_3d);
  free(y_vertices_3d);
  free(z_vertices_3d);

  *num_vertices_per_cell = xrealloc(*num_vertices_per_cell, part_num_cells *
                                    sizeof(**num_vertices_per_cell));
  *cell_to_vertex = xrealloc(*cell_to_vertex, num_cell_vertex_dependencies *
                             sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

static void generate_cube_grid(
  int comm_rank, int comm_size, int comp_id, int * grid_id,
  struct field_config * in_config, struct field_config * out_config) {

  unsigned n = 50;

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_core_mask;
  int * corner_core_mask;
  int * global_cell_id;
  int * global_corner_id;

  generate_part_cube_grid_information(n, &nbr_vertices, &nbr_cells,
                                      &num_vertices_per_cell, &cell_to_vertex,
                                      &x_vertices, &y_vertices, &x_cells,
                                      &y_cells, &global_cell_id,
                                      &cell_core_mask, &global_corner_id,
                                      &corner_core_mask, comm_rank, comm_size);

  yac_cdef_grid_unstruct(
    "cube", nbr_vertices, nbr_cells, (int*)num_vertices_per_cell,
    x_vertices, y_vertices, (int*)cell_to_vertex, grid_id);

  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, *grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, *grid_id);
  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, *grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, *grid_id);

  int corner_point_id, cell_point_id;
  yac_cdef_points_unstruct(
    *grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    &corner_point_id);
  yac_cdef_points_unstruct(
    *grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    &cell_point_id);

  yac_cdef_field("cube_to_icon", comp_id, &cell_point_id, 1, &(out_config->id));
  yac_cdef_field("icon_to_cube", comp_id, &cell_point_id, 1, &(in_config->id));

  out_config->data = xmalloc(nbr_cells * sizeof(*(out_config->data)));
  for (int i = 0; i < nbr_cells; ++i)
    out_config->data[i] =
      (cell_core_mask[i])?
        (test_harmonic(x_cells[i]*YAC_RAD, y_cells[i]*YAC_RAD)):DUMMY_VALUE;
  in_config->data = xmalloc(nbr_cells * sizeof(*(in_config->data)));
  out_config->data_size = nbr_cells;
  in_config->data_size = nbr_cells;
  free(cell_core_mask);
  free(corner_core_mask);
  free(global_cell_id);
  free(global_corner_id);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(num_vertices_per_cell);
  free(cell_to_vertex);
}

static void init_in_field_data(struct field_config field_config) {
  for (size_t i = 0; i < field_config.data_size; ++i)
    field_config.data[i] = DUMMY_VALUE;
}

static void check_results(
  struct field_config field_config, struct field_config ref_field_config) {

  if (field_config.data_size != ref_field_config.data_size) {
    fputs("ERROR(check_results): inconsistent data_size\n", stderr);
    exit(EXIT_FAILURE);
  }

  for (size_t i = 0; i < field_config.data_size; ++i) {
    if (fabs(field_config.data[i] - ref_field_config.data[i]) > 1e-3) {
      fputs("ERROR(check_results): data missmatch\n", stderr);
      exit(EXIT_FAILURE);
    }
  }
}
